import 'dart:collection';
import 'package:command/command.dart';
import 'package:meta/meta.dart';

typedef Command InsertCallback<E>(E item, int index);
typedef Command RemoveCallback<E>(E item);
typedef Command MoveItemCallback<E>(int oldIndex, int newIndex);
typedef Command ReplaceCallback<E>(int index, E replaceWith);

class CommandList<E> extends IterableBase<E> {
  CommandList({
    this.tag = "CommandList",
    Iterable<E> source,
    this.onInsert,
    this.onRemove,
    this.onMoveItem,
    this.onReplace,
  }) {
    if (source != null) _items.addAll(source);
  }

  List<E> _items = new List<E>();
  final String tag;
  final InsertCallback<E> onInsert;
  final RemoveCallback<E> onRemove;
  final MoveItemCallback<E> onMoveItem;
  final ReplaceCallback<E> onReplace;

  E operator [](int i) => _items[i];

  Iterator<E> get iterator => new CommandListIterator(_items);

  int indexOf(E element, [int start = 0]) {
    return _items.indexOf(element, start);
  }

  Command insert(E item, {int index}) {
    return _generateCommand(
      actions: new InsertItem(this, item, index),
      child: onInsert == null ? null : onInsert(item, index),
    );
  }

  Command remove(E item) {
    return _generateCommand(
      actions: new RemoveItem(this, item),
      child: onRemove == null ? null : onRemove(item),
    );
  }

  Command moveItem(int oldIndex, int newIndex) {
    return _generateCommand(
        actions: new MoveItem(this,oldIndex,newIndex),
        child: onMoveItem == null ? null : onMoveItem(oldIndex,newIndex),
    );
  }

  Command replace(int index, E replaceWith) {
    return _generateCommand(
        actions: new ReplaceItem(this, index, replaceWith),
        child: onReplace == null ? null : onReplace(index,replaceWith),
    );
  }

  Command _generateCommand({@required CommandAction actions,@required Command child}) {
    return new Command(actions,child: child)..execute();
  }
}

class CommandListIterator<E> extends Iterator<E> {
  List<E> _list;
  int _index = 0;
  E _current;

  CommandListIterator(List<E> list) {
    _list = list;
  }

  bool moveNext() {
    if (_index <= _list.length - 1) {
      _current = _list[_index];
      _index++;
      return true;
    } else {
      _current = null;
      return false;
    }
  }

  E get current => _current;
}

class InsertItem<E> implements CommandAction {
  final CommandList list;
  final E item;
  final int index;
  String get key => "${list.tag}.Insert";

  InsertItem(this.list, this.item, this.index);

  action() {
    if (index == null) {
      list._items.add(item);
    } else {
      list._items.insert(index, item);
    }
  }

  undoAction() {
    list._items.remove(item);
  }
}

class RemoveItem<E> implements CommandAction {
  final CommandList list;
  final E item;
  int _index;
  String get key => "${list.tag}.Remove";

  RemoveItem(this.list, this.item);

  action() {
    _index = getIndex();
    list._items.remove(item);
  }

  undoAction() {
    if (_index >= 0) {
      if (_index < list.length) {
        list._items.insert(_index, item);
      } else {
        list._items.add(item);
      }
    }
  }

  int getIndex() {
    int i = 0;
    for (var currentItem in list._items) {
      if (currentItem == item) {
        return i;
      }
      i++;
    }
    return -1;
  }
}

class MoveItem implements CommandAction {
  final CommandList list;
  final int oldIndex;
  final int newIndex;
  String get key => "${list.tag}.Move";

  MoveItem(this.list, this.oldIndex, this.newIndex);

  action() {
    var item = list._items.removeAt(oldIndex);
    list._items.insert(newIndex, item);
  }

  undoAction() {
    var item = list._items.removeAt(newIndex);
    list._items.insert(oldIndex, item);
  }
}

class ReplaceItem<E> implements CommandAction {
  final CommandList list;
  final int index;
  final E replaceWith;
  final E oldItem;
  String get key => "${list.tag}.Replace";

  ReplaceItem(this.list, this.index, this.replaceWith) : oldItem = list[index];

  action() {
    list._items[index] = replaceWith;
  }

  undoAction() {
    list._items[index] = oldItem;
  }
}
